/*
 * qwine-parse.cc: Parse the output of wine's autoctl channel and
 * create external windows to monitor the progress.  Currently only
 * implements a progress dialog.
 *
 * John R. Sheets <jsheets@codeweavers.com>
 *
 * To compile:
 *      moc -o mqwine-parse.h qwine-parse.h
 *      gcc -o qwine-parse qwine-parse.cc -lqt -I/usr/lib/qt2/include
 *
 * To run (in csh):
 *      ./wine -debugmsg +autoctl calc.exe |& ./qwine-parse
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include "progress-gui.moc.h"

// Progress tracing messages monitored by progress-gui
#define MSG_MODULE_LOADING    "module:MODULE_InitDll loading"
#define MSG_LIBRARY_COUNT1    "module:MODULE_DllProcessAttach found"
#define MSG_LIBRARY_COUNT2    "dependent DLLs for module "
#define MSG_EXEC_NAME         "module:make_lpCommandLine_name "
#define MSG_WINE_DONE         "global:MAIN_EmulatorRun done!"

bool g_noProgressMsg;    // Indicates whether progress messages should be shown
QString g_titleBar;      // Indicates the title bar of the progress dialog
QString g_logFileName;   // log file name

ProgressGUI::ProgressGUI(const QString &labelText,
                         QTextIStream *istr)
{
    m_istr = istr;
    m_progressDlg = new TimerProgressDialog (labelText);
    m_progressDlg->show ();
}

void ProgressGUI::handleInput ()
{
    QString line;	// One line of input from stdin
    // indicates whether the progress dialog has been closed
    static bool noMoreProgress = false;

    if (!m_istr)
        return;

    // readLine blocks the execution if there is no input coming
    // from the console. We should use select function to pool the
    // stdin first to make sure it won't block the UI.
    fd_set rfds;
    struct timeval tv;
    int retval;

    FD_ZERO (&rfds);
    FD_SET (0, &rfds);
    tv.tv_sec = tv.tv_usec = 0;
    retval = select (1, &rfds, NULL, NULL, &tv);
    if (!retval) return;

    line = m_istr->readLine ();

    if (line.isNull ())
    {
        int c;
        if ((c = getc(stdin)) != EOF)
        {
            // This is just an empty line.
            ungetc (c, stdin);
            printf ("\n");
            return;
        }
        else
        {
            // There is no more input coming from the pipe which means WINE has
            // exited unexpectedly.  Quit progress dialog too.
            printf ("\n");
            // Dump all trace messages to a log file
            if (!g_logFileName.isEmpty ()) {
	      QFile logFile(g_logFileName);
	      if (logFile.open (IO_WriteOnly))
		{
		  logFile.writeBlock (m_messageLog.ascii (),
				      m_messageLog.length ());
		  logFile.close ();
		}
	      else
                printf ("Unable to create the log file\n");
	    }

            QApplication::exit ();
            return;
        }
    }

    m_messageLog += line + "\n";
    if (!line.contains("progress:"))
    {	
        printf ("%s\n", (const char *) line);
        return;
    }
    else
    {
        // Surpress progress message is -noprogress option is turned on
        if (!g_noProgressMsg)
            printf ("%s\n", (const char *) line);

        if (line.contains (MSG_WINE_DONE))
        {
            if (!g_logFileName.isEmpty ())
            {
                // Dump all trace messages to a log file
                QFile logFile(g_logFileName);
                if (logFile.open (IO_WriteOnly))
                {
                    logFile.writeBlock (m_messageLog.ascii (),
                                        m_messageLog.length ());
                    logFile.close ();
                }
                else
                    printf ("Unable to create the log file\n");
            }
            // Quit the program when we saw this message
            QApplication::exit ();
        }

        // Let the progress dialog handle the progress messages
        if (!noMoreProgress)
            noMoreProgress = m_progressDlg->handleInput (line);
    }

}

TimerProgressDialog::TimerProgressDialog(const QString & labelText)
                    :QSemiModal (NULL, "Progress", TRUE)
{
    setCaption (labelText);
    resize (360, 160);
	
    m_progressBarFont = new QProgressBar (this);
    m_progressBarFont->hide ();
    m_progressBarModule = new QProgressBar (this);
    m_progressBarModule->setGeometry (40, 60, 280, 20);
    m_progressBarModule->setTotalSteps (100);
    m_progressBarModule->setProgress (0);
    m_label = new QLabel ("Initializing...", this);
    m_label->setAutoResize (true);
    m_label->move (40, 30);
    m_label2 = new QLabel ("", this);
    m_label2->setAutoResize (true);
    m_label2->move (40, 90);
    m_label2->hide ();
}

void TimerProgressDialog::doCreate (const QString &command)
{
    char *title;
    int start_range;
    int end_range;
    int start_val;
    int vals_read;

    // Parse out 'create' command, e.g.
    //    "create 0 754 1 *Building font metrics..."
    //
    vals_read = sscanf (command.ascii (), "create %d %d %d",
                        &start_range, &end_range, &start_val);
    if (vals_read < 3)
        printf ("ERROR reading create parameters!!\n");

    // Extract title for progress dialog
    // 
    title = strchr (command.ascii (), '*') + 1;

    m_progressBarFont->setTotalSteps (end_range - start_range);
    m_progressBarFont->setProgress (start_val);
    m_progressBarFont->setGeometry (40, 120, 280, 20);
	
    m_progressBarFont->show ();
    m_label2->setText (title);
    m_label2->show ();
}

void TimerProgressDialog::doIncrement (const QString &command)
{
    static int count = 0;
    // Chop off the number at the end, e.g.
    //     "progressdialog: increment 1"
    // 
    int chunk = command.length () - command.findRev (' ');
    QString strnum = command.right (chunk);
    count ++;

    m_progressBarFont->setProgress (m_progressBarFont->progress () + 
                                    strnum.toInt ());
}

bool TimerProgressDialog::handleInput (const QString &line)
{
    static QString targetName;
    static int moduleCount = -1;
    int index;

    // Increment the progress bar whenever a dynamic library is loaded
    if ((index = line.find (MSG_MODULE_LOADING)) >= 0)
    {
        QString moduleName = line.mid (index + strlen (MSG_MODULE_LOADING));
        if (moduleName.right (4) == ".dll" || moduleName.right (4) == ".DLL" ||
            moduleName.right (4) == ".exe" || moduleName.right (4) == ".EXE" ||
            moduleName.right (4) == ".drv" || moduleName.right (4) == ".DRV")
            moduleName = moduleName.left (moduleName.length () - 4);
            m_label->setText ("Loading library" + moduleName + "...");
        if (moduleCount == -1)
        {
            // At the beginning, we don't know how many modules
            // will be loaded.  So we just increment the progress
            // bar by a certain amount.
            m_progressBarModule->setProgress (
                    m_progressBarModule->progress () + 5);
        }
        else
        {
            if (m_progressBarModule-> progress () < 
                m_progressBarModule-> totalSteps () && moduleCount > 0) 
            {
                int increment = (m_progressBarModule->totalSteps () - 
                            m_progressBarModule-> progress ()) / moduleCount;
                moduleCount --;
                m_progressBarModule->setProgress (
                        m_progressBarModule->progress () + increment);
            }
        }
    }

    if (line.contains (MSG_LIBRARY_COUNT1) &&
        line.contains (MSG_LIBRARY_COUNT2))
    {
        // Extract the number of dependent dlls for the
        // current executable file.
        if (sscanf (line.ascii (),
           "progress:" MSG_LIBRARY_COUNT1 " %d " MSG_LIBRARY_COUNT2, 
            &moduleCount) != 1)
           moduleCount = -2;		//Error in parsing
    }

    // Extract the executable file name from the progress tracing messages
    if ((index = line.find (MSG_EXEC_NAME)) >= 0)
    {
        targetName = line.mid (index + strlen (MSG_EXEC_NAME));
    }

    if (!targetName.isEmpty () &&
        line.contains (MSG_MODULE_LOADING " " + targetName))
    {
        // The process of the executable file has been
        // attached.  We can assume that WINE has finished
        // loading the necessary libraries to run the
        // executable file.
        moduleCount = -3;
        delete this;
        return true;
    }

    // Process the progress dialog messages (coming from font metric building)
    if (line.contains ("progressdialog:"))
    {
        int index;
        QString command;
        QString title;

        // Handle 'create' command
        if ((index = line.find ("create")) >= 0)
        {
            command = line.right (line.length () - index);
            doCreate (command);
        }

        // Handle 'increment' command
        if ((index = line.find ("increment")) >= 0)
        {
            command = line.right (line.length () - index);
            doIncrement (command);
        }

        // Handle 'done' command
        if ((index = line.find ("done")) >= 0)
        {
            command = line.right (line.length () - index);
            m_progressBarFont->hide ();
            m_label2->hide ();
        }
    }
    return false;
}

void TimerProgressDialog::closeEvent (QCloseEvent *e)
{
    e->ignore ();
}

int main( int argc, char **argv )
{
    if (isatty(0))
        //The stdin is not from a pipe.  Simply exit the program
        exit (0);

    QString titleBarStr = "WINE";
    g_noProgressMsg = false;

    // Process command line arguments
    for (int i = 1; i < argc; i ++)
    {
        if (strcmp (argv[i], "-noprogress") == 0)
            g_noProgressMsg = true;
        if (strcmp (argv[i], "-title") == 0 && i + 1 < argc)
            titleBarStr = argv[i + 1];
        if (strcmp (argv[i], "-logfile") == 0 &&
            i + 1 < argc && argv[i+1][0] != '-')
            g_logFileName = argv[i + 1];
    }

    QApplication a( argc, argv );
    QTextIStream indata (stdin);
    ProgressGUI progressGui (titleBarStr, &indata);

    // Set up a timer to poll stdin.  This probably isn't the best way
    // to handle this, but it works pretty well.  The progress bar
    // might update a little more crisply with select() polling, but it
    // probably isn't worth the extra complexity for a transient dialog
    // like this.
    // 
    QTimer *timer = new QTimer (&progressGui);
    QObject::connect (timer, SIGNAL(timeout()), &progressGui, SLOT(handleInput()));
    timer->start (10);  // 100 ticks per second

    a.setMainWidget(&progressGui);
    return a.exec();
}
